use std::{
    io::Read,
    ops::Deref,
    sync::{
        mpsc::{channel, Sender},
        Arc,
    },
    thread::{self, JoinHandle},
    time::Duration,
};

use hashbrown::HashMap;
use parking_lot::{Mutex, RwLock};
use serialport::SerialPort;

use crate::error::{Error, SResult};

/// Handler of the serial port
pub struct SerialManager {
    serial_map: RwLock<HashMap<String, SerialQueue>>,
}

/// Queue data to allow non-blocking writes and blocking reads
#[derive(Debug)]
struct SerialQueue {
    _handle: JoinHandle<()>,
    sender: Sender<Vec<u8>>,
    serial: Arc<Mutex<Box<dyn SerialPort>>>,
}

impl SerialManager {
    pub fn new() -> Self {
        Self {
            serial_map: RwLock::new(HashMap::new()),
        }
    }

    /// Open the serial port with the given `port` device path and `baudrate`
    /// and bind it to the `name` key
    pub fn open(&mut self, name: &str, port: &str, baudrate: u32) -> SResult<()> {
        self.serial_map
            .write()
            .insert(name.to_owned(), SerialQueue::open(port, baudrate)?);
        Ok(())
    }

    /// Close the serial port if it is open
    pub fn close(&mut self, name: &str) -> SResult<()> {
        if self.serial_map.read().contains_key(name) {
            self.serial_map.write().remove(name);
            Ok(())
        } else {
            Err(Error::SerialNotOpen)
        }
    }

    /// Read `n` bytes from the serial port
    pub fn read_n_bytes(&self, name: &str, n: usize) -> SResult<Vec<u8>> {
        let map = self.serial_map.read();
        let port = map.get(name).ok_or(Error::SerialNotOpen)?;
        let mut buf = vec![0; n];
        port.lock().read_exact(&mut buf)?;
        Ok(buf)
    }

    /// Write `data` to the serial port
    pub fn enqueue_bytes(&self, name: &str, data: &[u8]) -> SResult<()> {
        let map = self.serial_map.read();
        map.get(name)
            .ok_or(Error::SerialNotOpen)?
            .enqueue_bytes(data)
    }
}

impl Deref for SerialQueue {
    type Target = Mutex<Box<dyn SerialPort>>;

    fn deref(&self) -> &Self::Target {
        &self.serial
    }
}

impl SerialQueue {
    /// Open the serial port and start a thread to write data as soon as it is
    /// available (by using a channel)
    fn open(port: &str, baudrate: u32) -> SResult<Self> {
        let serial = Arc::new(Mutex::new(
            serialport::new(port, baudrate)
                .timeout(Duration::from_millis(200))
                .open()?,
        ));
        let (tx, rx) = channel::<Vec<u8>>();
        let ser = Arc::clone(&serial);
        let _handle = thread::spawn(move || {
            while let Ok(data) = rx.recv() {
                ser.lock().write_all(&data).unwrap();
            }
        });
        Ok(Self {
            _handle,
            sender: tx,
            serial,
        })
    }

    /// Enqueue `data` to be written to the serial port, this is non-blocking
    fn enqueue_bytes(&self, data: &[u8]) -> SResult<()> {
        self.sender.send(data.to_vec())?;
        Ok(())
    }
}
